package be.iminds.ilabt.jfed.restapi.debugdata;

import be.iminds.ilabt.jfed.restapi.dao.DebugDataDao;
import be.iminds.ilabt.jfed.restapi.dropwizard.JFedLoggingAppender;
import be.iminds.ilabt.jfed.restapi.json.Experiment;
import ch.qos.logback.classic.spi.ILoggingEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;

import java.util.*;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * DebugDataManager:
 *   - This listens to JFedLoggingAppender to add logLines to the appropriate experimentDD's
 *   - todo This listens to jfed client calls to add ClientApiCallDD to the appropriate experimentDD's
 *   - This works a a cache for the ExperimentDD info in the DB
 */
public class DebugDataManager {
    private static final Logger LOG = LoggerFactory.getLogger(DebugDataManager.class);

    private final JFedLoggingAppender jFedLoggingAppender;
    private final DebugDataDao debugDataDao;

    public DebugDataManager(JFedLoggingAppender jFedLoggingAppender, DebugDataDao debugDataDao) {
        this.jFedLoggingAppender = jFedLoggingAppender;
        this.debugDataDao = debugDataDao;

        initAutoSave();
    }

    private final Map<Long, ExperimentDDBuilder> experimentDDBuildersById = new HashMap<>();
    public void addExperiment(ExperimentDDBuilder experimentDDBuilder) {
        assert experimentDDBuilder != null;
        assert experimentDDBuilder.getId() != null;
        LOG.debug("addExperiment("+experimentDDBuilder.getId()+")");
        experimentDDBuildersById.put(experimentDDBuilder.getId(), experimentDDBuilder);
        debugDataDao.insert(experimentDDBuilder.create());
    }
    public ExperimentDDBuilder getExperiment(Experiment experiment) {
        assert experiment != null;
        assert experiment.getId() != null;
        ExperimentDDBuilder res = getExperimentDDBuilder(experiment.getId());
        if (res == null) {
            //fallback: create it.
            LOG.error("ExperimentDD with id "+experiment.getId()+" was not cached and is not in the DB! Fallback: Will create new ExperimentDD.");
            res = new ExperimentDDBuilder(experiment);
            experimentDDBuildersById.put(res.getId(), res);
            debugDataDao.insert(res.create());
        }
        LOG.debug("getExperiment("+experiment.getId()+") res="+debugPrintNullHelper(res));
        return res;
    }

    public ExperimentDDBuilder getExperimentDDBuilder(long experimentId) {
        ExperimentDDBuilder res = experimentDDBuildersById.get(experimentId);
        if (res == null) {
            //get from DB
            res = debugDataDao.getExperimentDD(experimentId);
        }
        //todo for now, only uses cache, not DB
        LOG.debug("getExperimentDDBuilder("+experimentId+") res="+debugPrintNullHelper(res));
        return res;
    }
    public ExperimentDD getExperimentDD(long experimentId) {
        ExperimentDDBuilder res = getExperimentDDBuilder(experimentId);
        LOG.debug("getExperimentDD("+experimentId+") res="+debugPrintNullHelper(res));
        LOG.debug("There are ExperimentDD's for IDs: "+experimentDDBuildersById.keySet());
        if (res == null)
            return null;
        else
            return res.create();
    }

    public void save(ExperimentDDBuilder experimentDDBuilder) {
        assert experimentDDBuilder != null;
        assert experimentDDBuilder.getId() != null;
        LOG.debug("save("+experimentDDBuilder.getId()+")");
        processLogLines();

        ExperimentDD experimentDD = experimentDDBuilder.create();

        debugDataDao.update(experimentDD);
    }

    private final static int AUTO_SAVE_DEADLINE_MINUTES = 10;
    private final AtomicInteger nextServerResourceMethodCallId = new AtomicInteger();
    private final Map<Integer, ServerResourceMethodDDBuilder> serverResourceMethodIdMap = new HashMap<>();
    private final Map<Long, Date> experimentAutoSaveDeadlines = new HashMap<>();
    public void registerCallStart(ServerResourceMethodDDBuilder serverResourceMethodDD) {
        assert serverResourceMethodDD != null;
        int id = nextServerResourceMethodCallId.getAndIncrement();
        LOG.debug("registerCallStart("+id+")");
        serverResourceMethodIdMap.put(id, serverResourceMethodDD);
        MDC.put("serverResourceMethodCallId", id+"");
    }
    public void registerCallEnd(ServerResourceMethodDDBuilder serverResourceMethodDD, ExperimentDDBuilder experimentDD) {
        assert serverResourceMethodDD != null;
        LOG.debug("registerCallEnd()");
        //do not remove: all calls should override MDC,
        // if anything later uses this, it might belong to this call, so we log it as such
//        MDC.remove("serverResourceMethodCallId");

        if (experimentDD != null && experimentDD.getId() != null) {
            save(experimentDD);
            Date deadline = new Date(System.currentTimeMillis() + (AUTO_SAVE_DEADLINE_MINUTES * 60 * 1000l));
            experimentAutoSaveDeadlines.put(experimentDD.getId(), deadline);
        } else {
            LOG.warn("registerCallEnd() null or missing ID in experimentDD="+experimentDD);
        }
    }

    public void processLogLines() {
        if (jFedLoggingAppender == null) {
            LOG.debug("skipping processLogLines()");
            return;
        }

        List<ILoggingEvent> curLogLines = new ArrayList<>();
        jFedLoggingAppender.drainTo(curLogLines);
        LOG.debug("processLogLines() processes "+curLogLines.size()+" lines");
        for (ILoggingEvent loggingEvent : curLogLines) {
            String serverResourceMethodCallId = loggingEvent.getMDCPropertyMap().get("serverResourceMethodCallId");
            String experimentId = loggingEvent.getMDCPropertyMap().get("experimentId");
            if (serverResourceMethodCallId != null) {
                int id = Integer.parseInt(serverResourceMethodCallId);
                ServerResourceMethodDDBuilder builder = serverResourceMethodIdMap.get(id);
                if (builder != null) {
                    builder.addLogLine(new LogLineDDBuilder(loggingEvent));
                }
            }
//            if (experimentId != null) {
//                long id = Long.parseLong(experimentId);
//                ExperimentDDBuilder experimentDDBuilder = experimentDDBuildersById.get(id);
//                if (experimentDDBuilder != null) {
//                    //not logged seperately currently
//                }
//            }
        }
    }

    public void initAutoSave() {
        ScheduledExecutorService ses = Executors.newSingleThreadScheduledExecutor();
        ses.scheduleWithFixedDelay(() -> doAutoSave(), 10, 10, TimeUnit.SECONDS);
    }


    public void doAutoSave() {
        //always good to do this
        processLogLines();

//        saving everything all the time is NOT a good idea. We need something to only save "dirty" things to the DB
//        and preferably when they will not change again later.
//          Solution: save 10 minutes after last MainResource call ended, and on server shutdown
        Date now = new Date();
        for (ExperimentDDBuilder experimentDDBuilder : experimentDDBuildersById.values()) {
            Date deadline = experimentAutoSaveDeadlines.get(experimentDDBuilder.getId());
            if (deadline != null && now.before(deadline)) {
                LOG.debug("autosaving ExperimentDD "+experimentDDBuilder.getId());
                //clear deadline before save: anything that changes during save will trigger new save later
                experimentAutoSaveDeadlines.remove(experimentDDBuilder.getId());
                save(experimentDDBuilder);
            }
        }
    }



    private static String debugPrintNullHelper(Object o) {
        if (o == null)
            return "null";
        else
            return "non-null";
    }
}
